-- All code written and maintained by Yewchi 
-- zyewchi@gmail.com

--- Initialization --
--
------- Flags --
--
local SPELLBOOK_CONTINUE = 0xF0
local SPELLBOOK_STOP = 0xF1
--
------ /Flags --

------- Table key references --
--
local SPELL_OBJ_ID = 1
local SPELL_OBJ_NAME = 2
local SPELL_OBJ_SUBTEXT = 3

local MAX_CHAT_MESSAGE_LEN = 255

local ALPHABET_SEARCH_FOUND_INDEX = 1
--
------ /Table key references --

------- Static vals --
--
--
------- /Static vals --

------- Return values --
--
--
------ /Return values --

------- AddOn Variables --
--
local spellsNameTbl
spellNamesAlphabetSearch = {}
--
------ /AddOn Variables --
--
-- /Initialization --


--- Funcs --
--
--- A brutish attempt at a spellName look-up system. Unneccessarily large
--- structure, but not to a degree to cause any noticable lag for 99% of users. (<10MB)
-------------- CreateAlphabetTable()
local function CreateAlphabetTable()
	spellNamesAlphabetSearch = {}
	
	for spellName, _ in pairs(spellsNameTbl) do
		local thisTblOfTbls = spellNamesAlphabetSearch
		local spellNameLen = spellName:len()
		
		for i=1,spellNameLen,1 do
			local thisChar = spellName:sub(i,i) -- chars from string subs / string.byte() have the same size, even when used as a key
			
			if thisTblOfTbls[thisChar] == nil then
				thisTblOfTbls[thisChar] = {}
			end
			thisTblOfTbls = thisTblOfTbls[thisChar]
		end
		thisTblOfTbls[ALPHABET_SEARCH_FOUND_INDEX] = true
	end
end

local function SearchAlphabetTable(message, linkEscapePrefix)
	local LongestDepthFound = 0
	local LongestNameFoundLocation = {}
	local messageLength = message:len()
	local spellPrefix = ""
	
	local q = 1
	while q <= messageLength do -- O(n^2) (n < 256)
		local thisTblOfTbls = spellNamesAlphabetSearch
		local thisDepth = 0
		
		if linkEscapePrefix ~= nil and linkEscapePrefix ~= "" then
			local numCharsProcessed = q - 1

			q = message:sub(q, messageLength):find(FLEXI.SanitizeSpecialCharacters(linkEscapePrefix))
			if type(q) ~= "number" then
				break
			else
				q = min(messageLength, q + numCharsProcessed + linkEscapePrefix:len())
			end
		end
		
		for r=q,messageLength,1 do
			local thisChar = message:sub(r,r)
			thisTblOfTbls = thisTblOfTbls[thisChar]
			
			if thisTblOfTbls ~= nil then
				thisDepth = thisDepth + 1
				
				if thisTblOfTbls[ALPHABET_SEARCH_FOUND_INDEX] == true then
					if thisDepth > LongestDepthFound then
						LongestDepthFound = thisDepth
						LongestNameFoundLocation[1] = r - (thisDepth-1)
						LongestNameFoundLocation[2] = r
					end
				end
			else
				break
			end
		end
		q = q + 1
	end

	if LongestDepthFound > 0 then
		return message:sub(LongestNameFoundLocation[1], LongestNameFoundLocation[2])
	end
	return nil
end


---- Returns the spellObj generated from the spell at the page and 
---- index location of the spellbook. SPELLBOOK_CONTINUE if a 
---- non-spell is found. SPELLBOOK_STOP if there are no more spells.
-------------- CreateNewSpellObjFromBook()
local function CreateNewSpellObjFromBook(page, index)
	local skillType, spellId = GetSpellBookItemInfo(page, index)
	
	if spellId ~= nil then
		local spellName = GetSpellInfo(spellId)
		local spellSubtext = GetSpellSubtext(spellId)
if FLEXI.VERBOSE and spellSubtext == nil or spellSubtest == "" then print(FLEXI.flexiSpecifier..FLEXI.debugSpecifier.."spellSubtext for "..FLEXI.Printable(spellName).." is "..FLEXI.Printable(spellSubtext)..".") end
		
		if spellName ~= nil then
			return {spellId, spellName, spellSubtext} -- Success
		end
	end
	
	if skillType ~= nil then
		return SPELLBOOK_CONTINUE
	end

	return SPELLBOOK_STOP
end

local function AppendSpellIdToName(spellObj)
	local spellId, spellName, spellSubtext = unpack(spellObj)
	
	assert(spellId)
	assert(spellName)
	
	if spellsNameTbl[spellName] == nil then
		spellsNameTbl[spellName] = {}
	end
	
	local tblOfThisSpellName = spellsNameTbl[spellName]
	local spellSubtextRankNumber
	
	if spellSubtext ~= nil then 
		spellSubtextRankNumber = tonumber(spellSubtext:match("Rank%s(.+)")) -- Note: number instead of string comparison of ranks is required, some spells are Rank 10+ 
	end
	
	if spellSubtextRankNumber ~= nil then
		for i=1,#tblOfThisSpellName,1 do
			local thisIndexesSubtext = tblOfThisSpellName[i][SPELL_OBJ_SUBTEXT]
			if thisIndexesSubtext == nil or spellSubtextRankNumber > tonumber(thisIndexesSubtext:match("Rank%s(.+)")) then
				table.insert(tblOfThisSpellName, i, spellObj)
				
				return
			end
		end
	end
	
	table.insert(tblOfThisSpellName, spellObj) -- Just insert as the last element
	
	return
end

function FLEXI.Printable(val)
	local t = type(val)
	if val == nil then return "nil"
	elseif t == "number" or t == "string" then return val
	elseif t == "boolean" then return (val and "true" or "false")
	elseif t == "table" then return "[tbl]"
	elseif t == "function" then return "[func]"
	elseif t == "userdata" then return "[ud]"
	elseif t == "thread" then return "[thread]"
	end
	return "[WTF]"
end

function FLEXI.GetChannelArgs(channelStr)
	local channelNumber = tonumber(channelStr)
	
	if channelNumber ~= nil then
		if select(2, GetChannelName(channelNumber)) ~= nil then
			return "CHANNEL", nil, channelNumber
		end
		return nil
	end
	
	if channelStr == "s" or string.upper(channelStr) == "SAY" then
		return "SAY"
	elseif channelStr == "y" or string.upper(channelStr) == "YELL" then
		return "YELL"
	elseif channelStr == "p" or string.upper(channelStr) == "PARTY" then
		return "PARTY"
	elseif channelStr == "g" or string.upper(channelStr) == "GUILD" then
		return "GUILD"
	elseif channelStr == "o" or string.upper(channelStr) == "OFFICER" then
		return "OFFICER"
	elseif channelStr == "r" or string.upper(channelStr) == "RAID" then
		return "RAID"
	elseif channelStr == "rw" then
		return "RAID_WARNING"
	elseif channelStr == "i" or string.upper(channelStr) == "INSTANCE" then
		return "INSTANCE_CHAT"
	end
	
	return nil
end

-- TODO: Before further usage, need to confirm values are correct. 
function FLEXI.SanitizeSpecialCharacters(str)
	return str:gsub("%%", "%%%%")
			:gsub("%.", "%%%.")
			:gsub("%[", "%%%[")
			:gsub("%]", "%%%]")
			:gsub("%(", "%%%(")
			:gsub("%)", "%%%)")
			:gsub("^%^", "%%%^")
			:gsub("%+", "%%%+")
			:gsub("%-", "%%%-")
			:gsub("%*", "%%%*")
			:gsub("%?", "%%%?")
			:gsub("%$$", "%%%$")
end	

function FLEXI.DeduceMostLikelySpellName(args, linkEscapePrefix)
	local fullMsg = ""
	for i=2,#args,1 do
		fullMsg = string.format("%s %s", fullMsg, args[i])
	end
	
	return SearchAlphabetTable(fullMsg, linkEscapePrefix)
end

--- Returns the info (rank-including) of the highest-rank spell known for the 
--- desired name, in the GetSpellInfo() format.
-------- FLEXI.GetHighestRankSpellKnown
function FLEXI.GetHighestRankSpellKnown(spellName)
	assert(spellsNameTbl)
	
	local allRanksOfSpell = spellsNameTbl[spellName]
	
	if allRanksOfSpell ~= nil then
		local thisSpellObj = allRanksOfSpell[1]
		
		return thisSpellObj[SPELL_OBJ_NAME], thisSpellObj[SPELL_OBJ_SUBTEXT], _, _, _, _, thisSpellObj[SPELL_OBJ_ID]
	end
	
	return nil
end

function FLEXI.PopulateMySpellData()
	spellsNameTbl = {}
	
	local thisSpellObj
	
	local i = 1
	while thisSpellObj ~= SPELLBOOK_STOP do
		local thisSpellObj = CreateNewSpellObjFromBook(i, "") -- read "Arguments" - https://wowwiki.fandom.com/wiki/API_GetSpellBookItemInfo
		
		if type(thisSpellObj) == "table" then
			AppendSpellIdToName(thisSpellObj)
		elseif thisSpellObj == SPELLBOOK_STOP then
			break
		end
		
		i = i + 1
	end
	
	CreateAlphabetTable()
end
--
-- /Funcs --